# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER dotty syntax output.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import dispatch
from efilter import ast
from efilter import syntax
from efilter import query as q

from efilter.parsers.experiments import dotty


def __build_operator_lookup(*tables):
    lookup = {}
    for table in tables:
        for token, operator in table.iteritems():
            if not (isinstance(operator.handler, type) and
                    issubclass(operator.handler, ast.Expression)):
                continue

            lookup[operator.handler] = token

    return lookup


TOKENS = __build_operator_lookup(dotty.INFIX, dotty.PREFIX)


@dispatch.multimethod
def asdotty(expr):
    """Produces equivalent Dotty output to the AST.

    This class follows the visitor pattern. See documentation on VisitorEngine.
    """
    _ = expr
    raise NotImplementedError()


syntax.Syntax.register_formatter(shorthand="dotty", formatter=asdotty)


@asdotty.implementation(for_type=q.Query)
def asdotty(query):
    return asdotty(query.root)


@asdotty.implementation(for_type=ast.Within)
def asdotty(expr):
    lhs = expr.lhs
    rhs = expr.rhs
    left = asdotty(lhs)
    right = asdotty(rhs)
    token = "."

    if not isinstance(expr.lhs, (ast.ValueExpression,
                                 ast.Within)):
        left = "(%s)" % left
        token = " where "

    if not isinstance(expr.rhs, (ast.ValueExpression,
                                 ast.Within)):
        right = "(%s)" % right
        token = " where "

    return token.join((left, right))


@asdotty.implementation(for_type=ast.Any)
def asdotty(expr):
    return "any %s" % asdotty(ast.Within(expr.lhs, expr.rhs))


@asdotty.implementation(for_type=ast.Each)
def asdotty(expr):
    return "each %s" % asdotty(ast.Within(expr.lhs, expr.rhs))


@asdotty.implementation(for_type=ast.Literal)
def asdotty(expr):
    return repr(expr.value)


@asdotty.implementation(for_type=ast.Binding)
def asdotty(expr):
    return expr.value


@asdotty.implementation(for_type=ast.Complement)
def asdotty(expr):
    child = expr.value
    if (isinstance(child, ast.Equivalence) and
            len(child.children) == 2):
        return "%s != %s" % (asdotty(child.children[0]),
                             asdotty(child.children[1]))

    if isinstance(child, (ast.Within, ast.Binding,
                          ast.Literal)):
        return "not %s" % asdotty(child)

    # Put parens around everything else to be safe.
    return "not (%s)" % asdotty(child)


@asdotty.implementation(for_type=ast.BinaryExpression)
def asdotty(expr):
    try:
        token = TOKENS[type(expr)]
        separator = " %s " % token
        return "%s%s%s" % (asdotty(expr.lhs), separator, asdotty(expr.rhs))
    except KeyError:
        return "<subexpression cannot be formatted as dotty>"


@asdotty.implementation(for_type=ast.VariadicExpression)
def asdotty(expr):
    try:
        token = TOKENS[type(expr)]
        separator = " %s " % token
        return separator.join(asdotty(x) for x in expr.children)
    except KeyError:
        return "<subexpression cannot be formatted as dotty>"



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER query type inference.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import dispatch
from efilter import errors
from efilter import ast
from efilter import protocol
from efilter import query as q

from efilter.protocols import associative
from efilter.protocols import boolean
from efilter.protocols import reflective


@dispatch.multimethod
def infer_type(expr, scope=None):
    """Determines the types of each subexpression and validates sanity."""
    _ = expr, scope
    raise NotImplementedError()


@infer_type.implementation(for_type=q.Query)
def infer_type(query, scope=None):
    try:
        return infer_type(query.root, scope)
    except errors.EfilterError as error:
        error.query = query.source
        raise


@infer_type.implementation(for_type=ast.Literal)
def infer_type(expr, scope):
    _ = scope
    return type(expr.value)


@infer_type.implementation(for_type=ast.Binding)
def infer_type(expr, scope):
    if scope:
        result = reflective.reflect(scope, expr.value)

        if result:
            return result

    return protocol.AnyType


@infer_type.implementation(for_type=ast.Complement)
def infer_type(expr, scope):
    t = infer_type(expr.value, scope)
    if not protocol.isa(t, boolean.IBoolean):
        raise errors.EfilterTypeError(root=expr,
                                      actual=t,
                                      expected=boolean.IBoolean)

    return bool


@infer_type.implementation(for_type=ast.IsInstance)
def infer_type(expr, scope):
    _ = expr, scope
    return bool


@infer_type.implementation(for_type=ast.BinaryExpression)
def infer_type(expr, scope):
    lhs_type = infer_type(expr.lhs, scope)
    if not (lhs_type is protocol.AnyType
            or protocol.isa(lhs_type, expr.type_signature[0])):
        raise errors.EfilterTypeError(root=expr.lhs,
                                      expected=expr.type_signature[0],
                                      actual=lhs_type)

    rhs_type = infer_type(expr.rhs, scope)
    if not (lhs_type is protocol.AnyType
            or protocol.isa(rhs_type, expr.type_signature[1])):
        raise errors.EfilterTypeError(root=expr.rhs,
                                      expected=expr.type_signature[1],
                                      actual=rhs_type)

    return expr.return_signature


@infer_type.implementation(for_type=ast.VariadicExpression)
def infer_type(expr, scope):
    for subexpr in expr.children:
        t = infer_type(subexpr, scope)
        if not (t is protocol.AnyType
                or protocol.isa(t, expr.type_signature)):
            raise errors.EfilterTypeError(root=subexpr,
                                          expected=expr.type_signature,
                                          actual=t)

    return expr.return_signature


@infer_type.implementation(for_type=ast.Within)
def infer_type(expr, scope):
    t = infer_type(expr.context, scope)
    if not (t is protocol.AnyType
            or protocol.isa(t, associative.IAssociative)):
        raise errors.EfilterTypeError(root=expr,
                                      actual=t,
                                      expected=associative.IAssociative)

    return infer_type(expr.expression, t)


@infer_type.implementation(for_type=ast.Any)
def infer_type(expr, scope):
    t = infer_type(expr.context, scope)
    if not protocol.isa(t, boolean.IBoolean):
        raise errors.EfilterTypeError(root=expr,
                                      actual=t,
                                      expected=boolean.IBoolean)

    return bool


@infer_type.implementation(for_type=ast.Each)
def infer_type(expr, scope):
    t = infer_type(expr, scope)
    if not protocol.isa(t, boolean.IBoolean):
        raise errors.EfilterTypeError(root=expr,
                                      actual=t,
                                      expected=boolean.IBoolean)

    return bool



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER rule-based query analyzer.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import collections

from efilter import dispatch
from efilter import query as q
from efilter import errors
from efilter import ast
from efilter import protocol

from efilter.protocols import iset


Analysis = collections.namedtuple("Analysis",
                                  ("symbols", "eq_indexables"))


@dispatch.multimethod
def analyse(query, scope=None):
    """This is a rule-driven analyzer that gets a list of symbols and indexing.

    This class follows the visitor pattern. See documentation on VisitorEngine.

    The analyzer will produce a list of symbols required by the query (based on
    the Bindings/variables) and recommend a list of Bindings suitable for
    building an equivalence-based index (based on Equivalence expressions in
    the query).
    """
    _ = query, scope
    raise NotImplementedError()


@analyse.implementation(for_type=q.Query)
def analyse(query, scope=None):
    return analyse(query.root, scope)


@analyse.implementation(for_type=ast.Binding)
def analyse(expr, scope=None):
    _ = scope
    return Analysis((expr.value,), ())


@analyse.implementation(for_type=ast.Complement)
def analyse(expr, scope=None):
    return analyse(expr.value, scope)


@analyse.implementation(for_type=ast.Literal)
def analyse(expr, scope=None):
    _ = expr, scope
    return Analysis((), ())


@analyse.implementation(for_type=ast.BinaryExpression)
def analyse(expr, scope=None):
    lhsa = analyse(expr.lhs, scope)
    rhsa = analyse(expr.rhs, scope)

    return Analysis(
        iset.union(lhsa.symbols, rhsa.symbols),
        iset.union(lhsa.eq_indexables, rhsa.eq_indexables))


@analyse.implementation(for_type=ast.VariadicExpression)
def analyse(expr, scope=None):
    symbols = set()
    eq_indexables = set()

    for child in expr.children:
        analysis = analyse(child, scope)
        symbols.update(analysis.symbols)
        eq_indexables.update(analysis.eq_indexables)

    return Analysis(symbols, eq_indexables)


@analyse.implementation(for_type=ast.Within)
def analyse(expr, scope=None):
    if not isinstance(expr.lhs, ast.Binding):
        # Technically, the LHS context can be anything that implements
        # IAssociative, so a literal, or a subexpression that evaluates to
        # one are possible. Unfortunately, when that happens it is
        # non-trivial (read hard (read impossible)) to correctly determine
        # the scope for the RHS of the within-form.
        #
        # As this is the case, we are unable to create any hints, and
        # any symbols in the RHS expression are bound to an anonymous scope
        # and, as such, not useful.
        return analyse(ast.BinaryExpression(*expr), scope)

    scope = expr.lhs.value
    rhsa = analyse(expr.rhs, scope)
    symbols = set("%s.%s" % (scope, symbol) for symbol in rhsa.symbols)
    symbols.update(rhsa.symbols)
    symbols.add(expr.lhs.value)
    return rhsa._replace(symbols=symbols)


@analyse.implementation(for_type=ast.Membership)
def analyse(expr, scope=None):
    symbols = set()
    lha = analyse(expr.lhs, scope)
    rha = analyse(expr.rhs, scope)
    symbols.update(lha.symbols)
    symbols.update(rha.symbols)

    if (not isinstance(expr.rhs, ast.Literal)
            or not isinstance(expr.lhs, ast.Binding)):
        return Analysis(symbols, ())

    if not protocol.implements(expr.rhs.value, iset.ISet):
        # Yup, no can do.
        raise errors.EfilterTypeError(root=expr.rhs,
                                      actual=type(expr.rhs.value),
                                      expected=iset.ISet)

    return Analysis(symbols, (expr.lhs.value,))


@analyse.implementation(for_type=ast.Equivalence)
def analyse(expr, scope=None):
    literal = None
    indexables = set()
    symbols = set()
    for child in expr.children:
        if isinstance(child, ast.Literal):
            if literal is not None and literal != child.value:
                # This means something like 5 == 6 is being asked. This
                # expression will always be false and it makes no sense to
                # continue.
                return Analysis((), ())
            else:
                literal = child.value
        elif isinstance(child, ast.Binding):
            indexables.add(child.value)
            symbols.add(child.value)
        elif isinstance(child, ast.Within):
            # If we get a within-form, follow down as long as RHS is another
            # left form and the LHS is a binding. (something like
            # foo.bar.baz)
            within = child
            path = []
            while (isinstance(within, ast.Within)
                   and isinstance(within.lhs, ast.Binding)):
                path.append(within.lhs.value)
                symbols.add(".".join(path))
                within = within.rhs

            if isinstance(within, ast.Binding):
                path.append(within.value)

            remainder = analyse(child, scope)
            symbols.update(remainder.symbols)

            symbol = ".".join(path)
            symbols.add(symbol)
            indexables.add(symbol)
        else:
            analysis = analyse(child, scope)
            symbols.update(analysis.symbols)

    return Analysis(symbols, indexables)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER query wrapper.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


from efilter import syntax as s
from efilter import ast


def guess_source_syntax(source):
    if isinstance(source, ast.Expression):
        return "expression"

    if isinstance(source, basestring):
        return "dotty"

    if isinstance(source, tuple):
        return "lisp"

    return None


class Query(object):
    source = None
    root = None
    syntax = None
    application_delegate = None
    params = None

    def __init__(self, source, root=None, params=None, syntax=None,
                 application_delegate=None):
        super(Query, self).__init__()

        if isinstance(source, Query):
            # Run as a copy constructor with optional overrides.
            self.root = source.root
            self.source = source.source
            self.application_delegate = source.application_delegate
            self.syntax = source.syntax
            self.params = source.params
        elif isinstance(source, ast.Expression):
            # TODO: This will go away when other stops relying on it.
            self.root = source
        else:
            self.source = source

        # Override anything set by above code with explicit args.
        if syntax is not None:
            self.syntax = syntax

        if application_delegate is not None:
            self.application_delegate = application_delegate

        if params is not None:
            self.params = params

        if root is not None:
            if root != self.root:
                self.source = None  # No longer valid.
            self.root = root

        # Generate missing information.
        if not self.source and not self.root:
            raise ValueError("Must pass at least 'source' or 'root'.")

        if self.source and not self.root:
            # Run parser to generate AST.
            if not self.syntax:
                self.syntax = guess_source_syntax(self.source)

            parser_cls = s.Syntax.get_syntax(self.syntax)
            if not parser_cls:
                raise ValueError(
                    "Cannot find parser for syntax %r. Source was %r." %
                    (self.syntax, self.source))
            parser = parser_cls(original=self.source, params=self.params)
            self.root = parser.root
        elif self.root and not self.source:
            # Run formatter to generate the source.
            if not self.syntax:
                # Good, fully expressive default.
                self.syntax = "dotty"

            formatter = s.Syntax.get_formatter(self.syntax)
            if not formatter:
                # If we don't have a formatter for the explicit syntax, just
                # generate at least /something/.
                formatter = s.Syntax.get_formatter("dotty")
            self.source = formatter(self.root)

    def __str__(self):
        return unicode(self)

    def __unicode__(self):
        return unicode(self.source)

    def __repr__(self):
        return "Query(%s)" % repr(self.source)

    def __hash__(self):
        return hash(self.root)

    def __eq__(self, other):
        if not isinstance(other, Query):
            return False

        return self.root == other.root

    def __ne__(self, other):
        return not self.__eq__(other)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lisp-like EFILTER syntax.

This is mostly used in tests, in situations where dotty doesn't make it
obvious what the AST is going to look like, and manually setting up expression
classes is too verbose.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import ast
from efilter import syntax


EXPRESSIONS = {
    "var": ast.Binding,
    "!": ast.Complement,
    "map": ast.Map,
    "any": ast.Any,
    "each": ast.Each,
    "find": ast.Filter,
    "sort": ast.Sort,
    "in": ast.Membership,
    "regex": ast.RegexFilter,
    "|": ast.Union,
    "&": ast.Intersection,
    ">": ast.StrictOrderedSet,
    ">=": ast.PartialOrderedSet,
    "==": ast.Equivalence,
    "+": ast.Sum,
    "-": ast.Difference,
    "*": ast.Product,
    "/": ast.Quotient,
}


class Parser(syntax.Syntax):
    """Parses the lisp expression language into the query AST."""

    @property
    def root(self):
        return self._parse_atom(self.original)

    def _parse_atom(self, atom):
        if isinstance(atom, tuple):
            return self._parse_s_expression(atom)

        return ast.Literal(atom)

    def _parse_s_expression(self, atom):
        car = atom[0]
        cdr = atom[1:]

        # Bindings are a little special.
        if car == "var":
            return ast.Binding(cdr[0])

        return EXPRESSIONS[car](*(self._parse_atom(a) for a in cdr))


syntax.Syntax.register_parser(Parser, shorthand="lisp")



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER abstract type system.

The type protocols defined under efilter.protocols.* provide a very thin layer
over Python's builtin types, defined as collections of related functions with
defined semantics. Each type protocol is intended to uniformly support a
specific behavior across any type that participates in the protocol.

To participate in a protocol, two things are required:
1) Implementations of each of the member functions must be provided.
2) The type must be formally added to the protocol.

In this manner, we are able to declare strict compositional types on atoms and
expressions in the EFILTER AST and allow type hierarchies external to EFILTER
(Plaso Events, Rekall Entities) to be passed to the EFILTER transforms without
casting or wrapping.

The compositional, flat nature of the type protocols makes it simple to support
basic type inference, by annotating each expression type with sets of
protocols it requires on its children and guarantees on its return type.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import abc


class AnyType(object):
    """Sentinel used to provide a default implementation of a protocol.

    If you need to provide a default implementation of functions in a
    protocol (for example, providing fall-through behavior for objects that
    don't participate in the protocol) you may pass this type in place of
    'object'. This will cause the multimethod functions to fall through to
    this default implementation, but won't cause 'object' to be a subclass
    of the protocol.

    Example:
        MyProtocol.implement(for_type=AnyType,
                             implementations={foo=lambda x: "foo"})

        foo(5)  # => "foo"
        isinstance(5, MyProtocol)  # => False
        implements(5, MyProtocol)  # => True
    """


BUILTIN_TYPES = (int, float, long, complex, basestring, tuple, list, dict, set,
                 frozenset, type(None), AnyType)


def implements(obj, protocol):
    """Does the object 'obj' implement the 'prococol'?"""
    if isinstance(obj, type):
        raise TypeError("First argument to implements must be an instance. "
                        "Got %r." % obj)
    return isinstance(obj, protocol) or issubclass(AnyType, protocol)


def isa(cls, protocol):
    """Does the type 'cls' participate in the 'protocol'?"""
    if not isinstance(cls, type):
        raise TypeError("First argument to isa must be a type. Got %r." % cls)

    if not isinstance(protocol, type):
        raise TypeError(("Second argument to isa must be a type or a Protocol. "
                         "Got an instance of %r.") % type(protocol))
    return issubclass(cls, protocol) or issubclass(AnyType, protocol)


class Protocol(object):
    """Collection of related functions that operate on a type (interface)."""
    __metaclass__ = abc.ABCMeta

    _protocol_functions = set()

    @classmethod
    def functions(cls):
        result = set(cls._protocol_functions)

        for scls in cls.mro():
            protocol_functions = getattr(scls, "_protocol_functions", None)
            if protocol_functions:
                result.update(protocol_functions)

        return result

    @classmethod
    def implemented(cls, for_type):
        for function in cls.functions():
            if not function.implemented_for_type(for_type):
                raise TypeError(
                    "%r doesn't implement %r so it cannot participate in "
                    "the protocol %r." %
                    (for_type, function.func.func_name, cls))

        cls.register(for_type)

    @classmethod
    def _implement_for_type(cls, for_type, implementations):
        # AnyType is a sentinel that means the multimethod function should
        # just dispatch on 'object'.
        dispatch_type = object if for_type is AnyType else for_type
        protocol_functions = cls.functions()
        remaining = set(protocol_functions)

        for func, impl in implementations.iteritems():
            if func not in protocol_functions:
                func_name = getattr(func, "func_name", repr(func))
                raise TypeError("Function %s is not part of the protocol %r." %
                                (func_name, cls))

            func.implement(for_type=dispatch_type,
                           implementation=impl)
            remaining.remove(func)

        if remaining:
            raise TypeError(
                "%s.implement invokation must provide implementations of "
                "%r" % (cls.__name__, remaining))

        cls.implemented(for_type=for_type)

    @classmethod
    def implicit_static(cls, for_type):
        """Automatically generate implementations for a type.

        Implement the protocol for the 'for_type' type by dispatching each
        member function of the protocol to an instance method of the same name
        declared on the type 'for_type'.

        Arguments:
            for_type: The type to implictly implement the protocol with.

        Raises:
            TypeError if not all implementations are provided by 'for_type'.
        """
        implementations = {}
        for function in cls.functions():
            method = getattr(for_type, function.func_name, None)
            if not callable(method):
                raise TypeError(
                    "%s.implicit invokation on type %r is missing instance "
                    "method %r." % (cls.__name__, for_type, function.func_name))

            implementations[function] = method

        return cls.implement(for_type=for_type, implementations=implementations)

    @staticmethod
    def _build_late_dispatcher(func_name):
        """Return a function that calls method 'func_name' on objects.

        This is useful for building late-bound dynamic dispatch.

        Arguments:
            func_name: The name of the instance method that should be called.

        Returns:
            A function that takes an 'obj' parameter, followed by *args and
            returns the result of calling the instance method with the same
            name as the contents of 'func_name' on the 'obj' object with the
            arguments from *args.
        """
        def _late_dynamic_dispatcher(obj, *args):
            method = getattr(obj, func_name, None)
            if not callable(method):
                raise NotImplementedError(
                    "Instance method %r is not implemented by %r." % (
                        func_name, obj))

            return method(*args)

        return _late_dynamic_dispatcher

    @classmethod
    def implicit_dynamic(cls, for_type):
        """Automatically generate late dynamic dispatchers to type.

        This is similar to 'implicit_static', except instead of binding the
        instance methods, it generates a dispatcher that will call whatever
        instance method of the same name happens to be available at time of
        dispatch.

        This has the obvious advantage of supporting arbitrary subclasses, but
        can do no verification at bind time.

        Arguments:
            for_type: The type to implictly implement the protocol with.
        """
        implementations = {}
        for function in cls.functions():
            implementations[function] = cls._build_late_dispatcher(
                func_name=function.func_name)

        return cls.implement(for_type=for_type, implementations=implementations)

    @classmethod
    def implement(cls, implementations, for_type=None, for_types=None):
        """Provide protocol implementation for a type.

        Register all implementations of multimethod functions in this
        protocol and add the type into the abstract base class of the
        protocol.

        Arguments:
            implementations: A dict of (function, implementation), where each
                function is multimethod and each implementation is a callable.
            for_type: The concrete type implementations apply to.
            for_types: Same as for_type, but takes a tuple of types.

            You may not supply both for_type and for_types for obvious reasons.

        Raises:
            ValueError for arguments.
            TypeError if not all implementations are provided or if there
                are issues related to polymorphism (e.g. attempting to
                implement a non-multimethod function.
        """
        if for_type:
            if for_types:
                raise ValueError("Cannot pass both for_type and for_types.")
            for_types = (for_type,)
        elif for_types:
            if not isinstance(for_types, tuple):
                raise TypeError("for_types must be passed as a tuple of "
                                "types (classes).")
        else:
            raise ValueError("Must pass either for_type or for_types.")

        for type_ in for_types:
            cls._implement_for_type(for_type=type_,
                                    implementations=implementations)



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system.

The data provider is intended to be similar to a collector in Rekall. Using the
protocol, an application can expose generators of objects selected based on
different criteria, such as desired types and hinted queries.

This module is experimental and subject to frequent change.
"""

from efilter import dispatch
from efilter import protocol


# pylint: disable=unused-argument


@dispatch.multimethod
def provide(provider, query=None, params=None):
    raise NotImplementedError()


@dispatch.multimethod
def declare_output(provider, query=None):
    raise NotImplementedError()


@dispatch.multimethod
def declare_input(provider, query=None):
    raise NotImplementedError()


class IDataProvider(protocol.Protocol):
    _protocol_functions = (provide, declare_input, declare_output)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER abstract syntax.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


class EfilterError(Exception):
    query = None
    _root = None
    message = None
    start = None
    end = None

    def __init__(self, query=None, message=None, root=None, start=None,
                 end=None):
        super(EfilterError, self).__init__(message)

        self.query = query
        self.message = message
        self.root = root

        if start is not None:
            self.start = start

        if end is not None:
            self.end = end

    @property
    def root(self):
        return self._root

    @root.setter
    def root(self, value):
        self._root = value

        try:
            self.start = value.start
            self.end = value.end
        except AttributeError:
            self.start = None
            self.end = None

    @property
    def text(self):
        return self.message

    @property
    def source(self):
        if not self.query:
            return None

        if self.start is not None and self.end is not None:
            return "%s >>> %s <<< %s" % (
                self.query[0:self.start],
                self.query[self.start:self.end],
                self.query[self.end:])
        elif self.query:
            return self.query

    def __str__(self):
        return "%s (%s) in query %r" % (
            type(self).__name__,
            self.text,
            self.source)

    def __repr__(self):
        return "%s(message=%r, start=%r, end=%r)" % (
            type(self), self.message, self.start, self.end)


class EfilterNoneError(EfilterError):
    pass


class EfilterParseError(EfilterError):
    token = None

    def __init__(self, *args, **kwargs):
        self.token = kwargs.pop("token", None)
        super(EfilterParseError, self).__init__(*args, **kwargs)


class EfilterKeyError(EfilterError):
    key = None

    @property
    def text(self):
        if self.message:
            return self.message

        if self.key:
            return "No key %r." % self.key

        return None

    def __init__(self, *args, **kwargs):
        self.expected = kwargs.pop("key")

        super(EfilterKeyError, self).__init__(*args, **kwargs)


class EfilterTypeError(EfilterError):
    expected = None
    actual = None

    @property
    def text(self):
        if self.message:
            return self.message

        if self.expected and self.actual:
            return "Expected type %r, got %r instead." % (self.expected,
                                                          self.actual)

        return None

    def __init__(self, *args, **kwargs):
        self.expected = kwargs.pop("expected")
        self.actual = kwargs.pop("actual")

        super(EfilterTypeError, self).__init__(*args, **kwargs)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER individual object filter and matcher.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


# pylint: disable=function-redefined

import collections
import re

from efilter import ast
from efilter import dispatch
from efilter import errors
from efilter import protocol
from efilter import query as q


from efilter.protocols import associative
from efilter.protocols import iset
from efilter.protocols import reflective
from efilter.protocols import repeated


Result = collections.namedtuple("Result", ["value", "branch", "sort_key"])


@dispatch.multimethod
def solve(query, bindings):
    """Evaluate the 'query' using variables in 'bindings'.

    Canonical implementation of the EFILTER AST's actual behavior. This may
    not be the most optimal way of executing the query, but it is guaranteed
    to have full coverage without falling through to some other implementation.

    Arguments:
        query: The instance of Query to evaluate against data in bindings.
        bindings: An object implementing IAssociative (like a dict) containing
            pairs of variable -> value.

    Returns:
        Instance of Result, with members set as follows:

            value: The result of evaluation. The type of the result can be
                determined by calling infer_type on 'query'.

            branch: An instance of Expression, representing a subtree of 'query'
                that was that last branch evaluated before a match was produced.
                This only applies to simple queries using AND/OR and NOT
                operators, which evaluate to booleans and can terminate early.
                For other queries this will be set to None.

            sort_key: A key by which the 'bindings' should be sorted with
                respect to other groups of bindings of the same type. Not
                currently implemented.
    """
    _ = query, bindings
    raise NotImplementedError()


@solve.implementation(for_type=q.Query)
def solve(query, bindings):
    try:
        return solve(query.root, bindings)
    except errors.EfilterError as error:
        if not error.query:
            error.query = query.source
        raise


@solve.implementation(for_type=ast.Literal)
def solve(expr, bindings):
    """Returns just the value of literal."""
    _ = bindings
    return Result(expr.value, (), ())


@solve.implementation(for_type=ast.Binding)
def solve(expr, bindings):
    """Returns the value of the binding (var) named in the expression."""
    try:
        return Result(associative.resolve(bindings, expr.value), (), ())
    except (KeyError, AttributeError) as e:
        # Raise a better exception for accessing a non-existent member.
        raise errors.EfilterKeyError(root=expr, key=expr.value, message=e)
    except (TypeError, ValueError) as e:
        # Raise a better exception for what is probably a null pointer error.
        if isinstance(bindings, type(None)):
            raise errors.EfilterNoneError(
                root=expr,
                message="Trying to access member %r of a null." % expr.value)
        else:
            raise
    except NotImplementedError as e:
        raise errors.EfilterError(
            root=expr,
            message="Trying to access member %r of an instance of %r." %
            (expr.value, type(bindings)))


@solve.implementation(for_type=ast.Map)
def solve(expr, bindings):
    """Solves the map-form, by recursively calling its RHS with new bindings.

    let-forms are binary expressions. The LHS should evaluate to an IAssociative
    that can be used as new bindings with which to solve a new query, of which
    the RHS is the root. In most cases, the LHS will be a Binding (var).

    Typically, map-forms result from the dotty "dot" (.) operator. For example,
    the query "User.name" will translate to a map-form with the binding "User"
    on LHS and a binding to "name" on the RHS. With top-level bindings being
    something like {"User": {"name": "Bob"}}, the Binding on the LHS will
    evaluate to {"name": "Bob"}, which subdict will then be used on the RHS as
    new bindings, and that whole form will evaluate to "Bob".
    """
    lhs = solve(expr.lhs, bindings)

    try:
        values = []
        for value in repeated.getvalues(lhs.value):
            value_ = solve(expr.rhs, value)
            values.append(value_.value)
    except errors.EfilterNoneError as error:
        error.root = expr
        raise

    return Result(repeated.meld(*values), (), ())


@solve.implementation(for_type=ast.Filter)
def solve(expr, bindings):
    """Filter values on the LHS by evaluating RHS with each value.

    Returns any LHS values for which RHS evaluates to a true value.
    """
    lhs = solve(expr.lhs, bindings)

    results = []
    for value in repeated.getvalues(lhs.value):
        if solve(expr.rhs, value).value:
            results.append(value)

    return Result(repeated.meld(*results), (), ())


@solve.implementation(for_type=ast.Sort)
def sort(expr, bindings):
    """Sort values on the LHS by the value they yield when passed to RHS."""
    values = repeated.getvalues(solve(expr.lhs, bindings).value)
    values = sorted(values, key=lambda val: solve(expr.rhs, val))
    return Result(repeated.meld(*values), (), ())


@solve.implementation(for_type=ast.Each)
def solve(expr, bindings):
    """Return True if RHS evaluates to a true value with each state of LHS.

    If LHS evaluates to a normal IAssociative object then this is the same as
    a regular let-form, except the return value is always a boolean. If LHS
    evaluates to a repeared var (see efilter.protocols.repeated) of
    IAssociative objects then RHS will be evaluated with each state and True
    will be returned only if each result is true.
    """
    branch_bindings = solve(expr.lhs, bindings).value
    for state in repeated.getvalues(branch_bindings):
        result = solve(expr.rhs, state)
        if not result.value:
            return result

    return Result(True, (), ())


@solve.implementation(for_type=ast.Any)
def solve(expr, bindings):
    """Same as Each, except returning True on first true result at LHS."""
    branch_bindings = solve(expr.lhs, bindings).value
    result = Result(False, (), ())
    for state in repeated.getvalues(branch_bindings):
        result = solve(expr.rhs, state)
        if result.value:
            return result

    return result


@solve.implementation(for_type=ast.IsInstance)
def solve(expr, bindings):
    """Typecheck whether LHS is type on the RHS."""
    lhs = solve(expr.lhs, bindings)
    t = reflective.reflect(type(bindings), expr.rhs)
    return Result(protocol.implements(lhs.value, t), (), ())


@solve.implementation(for_type=ast.Complement)
def solve(expr, bindings):
    result = solve(expr.value, bindings)
    return result._replace(value=not result.value)


@solve.implementation(for_type=ast.Intersection)
def solve(expr, bindings):
    result = Result(False, (), ())
    for child in expr.children:
        result = solve(child, bindings)
        if not result.value:
            return result

    return result


@solve.implementation(for_type=ast.Union)
def solve(expr, bindings):
    for child in expr.children:
        result = solve(child, bindings)
        if result.value:
            return result._replace(branch=child)

    return Result(False, (), ())


@solve.implementation(for_type=ast.Sum)
def solve(expr, bindings):
    total = 0
    for child in expr.children:
        total += solve(child, bindings).value

    return Result(total, (), ())


@solve.implementation(for_type=ast.Difference)
def solve(expr, bindings):
    children = iter(expr.children)
    difference = solve(next(children), bindings).value
    for child in children:
        difference -= solve(child, bindings).value

    return Result(difference, (), ())


@solve.implementation(for_type=ast.Product)
def solve(expr, bindings):
    product = 1
    for child in expr.children:
        product *= solve(child, bindings).value

    return Result(product, (), ())


@solve.implementation(for_type=ast.Quotient)
def solve(expr, bindings):
    children = iter(expr.children)
    quotient = solve(next(children), bindings).value
    for child in children:
        quotient /= solve(child, bindings).value

    return Result(quotient, (), ())


@solve.implementation(for_type=ast.Equivalence)
def solve(expr, bindings):
    children = iter(expr.children)
    first_value = solve(next(children), bindings).value
    for child in children:
        if solve(child, bindings).value != first_value:
            return Result(False, (), ())

    return Result(True, (), ())


@solve.implementation(for_type=ast.Membership)
def solve(expr, bindings):
    element = solve(expr.element, bindings).value
    values = solve(expr.set, bindings).value
    return Result(element in values, (), ())


@solve.implementation(for_type=ast.RegexFilter)
def solve(expr, bindings):
    string = solve(expr.string, bindings).value
    pattern = solve(expr.regex, bindings).value

    return Result(re.compile(pattern).match(str(string)), (), ())


@solve.implementation(for_type=ast.ContainmentOrder)
def solve(expr, bindings):
    _ = bindings
    iterator = iter(expr.children)
    x = solve(next(iterator), bindings).value
    for y in iterator:
        y = solve(y, bindings).value
        if not iset.issubset(x, y):
            return Result(False, (), ())
        x = y

    return Result(True, (), ())


@solve.implementation(for_type=ast.StrictOrderedSet)
def solve(expr, bindings):
    iterator = iter(expr.children)
    min_ = solve(next(iterator), bindings).value

    if min_ is None:
        return Result(False, (), ())

    for child in iterator:
        val = solve(child, bindings).value

        if not min_ > val or val is None:
            return Result(False, (), ())

        min_ = val

    return Result(True, (), ())


@solve.implementation(for_type=ast.PartialOrderedSet)
def solve(expr, bindings):
    iterator = iter(expr.children)
    min_ = solve(next(iterator), bindings).value

    if min_ is None:
        return Result(False, (), ())

    for child in iterator:
        val = solve(child, bindings).value

        if min_ < val or val is None:
            return Result(False, (), ())

        min_ = val

    return Result(True, (), ())



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER type system.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter.protocols import indexable
from efilter.protocols import repeated


class ListRepetition(object):
    """Repeated variable backed by a list."""

    _delegate = None

    def __init__(self, first_value=None, *values):
        self._delegate = []

        if first_value is None:
            return

        self._value_type = repeated.value_type(first_value)
        self.add_value(first_value)

        for value in values:
            if repeated.value_type(value) != self.value_type():
                raise TypeError(
                    "All values of a repeated var must be the of same type."
                    " First argument was of type %r, but argument %r is of "
                    " type %r." %
                    (self.value_type(), value, repeated.value_type(value)))
            self.add_value(value)

    def add_value(self, value):
        """Add a value to this repeated var.

        WARNING: this mutates the object (it's NOT copy on write). Unless
        you're absolutely certain of what you're doing, you most likely want
        to call repeated.meld(x, y) instead.
        """
        self._delegate.extend(repeated.getvalues(value))

    def value_type(self):
        return self._value_type

    def getvalues(self):
        # Return a copy because delegate is mutable and we don't want things
        # to blow up.
        return self._delegate[:]

    def value_eq(self, other):
        if isinstance(other, type(self)):
            # pylint: disable=protected-access
            return sorted(self._delegate) == sorted(other._delegate)

        return sorted(self._delegate) == sorted(repeated.getvalues(other))

    def __eq__(self, other):
        if not isinstance(other, repeated.IRepeated):
            return False

        return self.value_eq(other)

    def __ne__(self, other):
        return not self == other

    def indices(self):
        result = set()
        for value in self._delegate:
            result |= frozenset(indexable.indices(value))

        return result

    def value_apply(self, f):
        return repeated.repeated(*[f(x) for x in self.getvalues()])

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__,
                           ", ".join([repr(x) for x in self.getvalues()]))


indexable.IIndexable.implicit_static(ListRepetition)
repeated.IRepeated.implicit_static(ListRepetition)
repeated.repeated.implement(for_type=object,
                            implementation=ListRepetition)



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER query hinter.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


from efilter import dispatch
from efilter import errors
from efilter import ast
from efilter import query as q

from efilter.transforms import normalize


@dispatch.multimethod
def hint(expr, selector):
    """Retargets the query to apply to a subexpression, to be used for hinting.

    Discussion of mechanism and rationale:
    ======================================

    Bear with me on this one.

    As with any database-like system, certain EFILTER queries can be satisfied
    by following several equally valid, but differently expensive strategies.
    Because traditional notions of cardinality and available indexing do not
    apply to most systems that provide EFILTER with data (and, at any rate,
    such concerns are usually negligible in comparison with the actual
    collection of data), it becomes useful for the expert system supplying the
    data to be able to influence the strategy based on what it knows about the
    relative trade-offs.

    To give a concrete example, take the following query:
        VAD.flags contains {'execute', 'write'} and VAD.process.name == 'init'

    (In this example a VAD is a virtual address descriptor, which is a data
    structure used by operating systems to keep track of pageable memory.)

    A naive expert (uh...) system will generate all possible VADs and feed them
    to the EFILTER solve function. An optimized system - lets call it an expert
    expert system - can only collect VADs belonging to the 'init' process at
    fraction of the cost, and supply those to EFILTER for indexing and
    filtering. This is possible, because the expert system knows that processes
    are the first to be collected, and that each process holds a reference to
    related VADs, making it beneficial to skip processes that don't apply. What
    the expert system needs is a way to signal that such a strategy is
    available, and a way to recognize the processes which cannot be skipped.

    Enter the Hinter - taking the example above, the expert system can run:
        q = Query("VAD.flags contains {'execute', 'write'} "
                  " and VAD.process.name == 'init'")
        h = hint(q, selector='VAD.process')

    This will cause the Hinter to generate a hint query equivalent to the
    expression "name == 'init'", which can be applied to a process object
    for prefiltering. Amazing.
    """
    _ = expr, selector
    raise NotImplementedError()


@hint.implementation(for_type=q.Query)
def hint(query, selector):
    parsed_selector = tuple(selector.split(".")) if selector else ()

    # Make sure the query is reasonably shaped with respect to let-forms.
    query = normalize.normalize(query)
    new_root = hint(query.root, parsed_selector)
    return q.Query(query, root=new_root)


@hint.implementation(for_type=ast.Within)
def hint(expr, selector):
    # Are we already in a branch that we're preserving?
    if not selector:
        return expr

    # Verify that the AST below is well-formed.
    if not isinstance(expr.lhs, ast.Binding):
        # Can't do anything - might not be correct - perhaps we should
        # blow up here? (Only within-forms with a Binding on the LHS can be
        # hinted.)
        #
        # TODO(adamsh): It's likely that a let with an LHS that's not a binding
        # has very limited uses anyway - maybe consider not allowing it in the
        # AST.
        raise errors.EfilterError(
            message=("Hinter can only optimize let forms where lhs is a "
                     "binding (a variable). Got %r instead.") % expr,
            root=expr)

    if selector[0] == expr.lhs.value:
        # Next part of the selector is the same as our LHS. Descend.
        return hint(expr.rhs, selector[1:])

    return None  # Eliminate this branch.


def _build_variadic(expr, selector):
    children = []
    for child in expr.children:
        branch = hint(child, selector)
        if branch:
            children.append(branch)

    return children


@hint.implementation(for_type=ast.VariadicExpression)
def hint(expr, selector):
    children = _build_variadic(expr, selector)
    if not children:
        return None

    if len(children) == 1:
        # We play it safe and keep the only child in case the return type
        # of this expression isn't just a boolean (math).
        # More specific visitors will eliminate the whole thing.
        return children[0]

    return type(expr)(*children)


@hint.implementation(for_type=ast.Relation)
def hint(expr, selector):
    children = _build_variadic(expr, selector)

    if len(children) == 1:
        return None  # It's pointless to execute this relation.

    return type(expr)(*children)


@hint.implementation(for_type=ast.Complement)
def hint(expr, selector):
    child = hint(expr.value, selector)
    if child is None:
        return None

    return type(expr)(child)


@hint.implementation(for_type=ast.IsInstance)
def hint(expr, selector):
    _ = selector
    return expr


@hint.implementation(for_type=ast.BinaryExpression)
def hint(expr, selector):
    lhs = hint(expr.lhs, selector)
    if not lhs:
        return None

    rhs = hint(expr.rhs, selector)
    if not rhs:
        return None

    return type(expr)(lhs, rhs)


@hint.implementation(for_type=ast.Literal)
def hint(expr, selector):
    _ = selector
    return expr


@hint.implementation(for_type=ast.Binding)
def hint(expr, selector):
    if not selector:
        return expr

    return None



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER versioning scheme.

EFILTER version is the UTC UNIX epoch of the git commit that the package is
being built from. This allows us to disambiguate different builds of the same
major version, and simply map each build back to the commit it came from.

It would be more convenient to just use the git commitish as version strings,
but PEP 0440 mandates that version numbers increment over time, which the
commitish, being output of a hash function, doesn't.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


import re

try:
    import datetime
    import pytz
    import subprocess

    # The below functionality is only available if dateutil is installed.
    from dateutil import parser

    def _unix_epoch(date):
        """Convert datetime object to a UTC UNIX timestamp."""
        td =  date - datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC)
        return int(td.total_seconds())

    def run_git_log():
        """Generate version based on date of last commit.

        Returns:
            UTC UNIX timestamp as int on success, or None.
        """
        try:
            p = subprocess.Popen(
                ["git", "log", "-1", "--format=%cd", "--date=iso-strict"],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
            p.stderr.close()
            output = p.stdout.readlines()[0]
            date = parser.parse(output)
            return _unix_epoch(date)
        except (OSError, IndexError):
            # Even if git log fails (because it's not in a git repo), the call
            # may still 'succeed' as far as subprocess.Popen is concerned,
            # hence the IndexError exception. I don't know why Python sometimes
            # ignores the return code.
            return None
except ImportError:
    # If there's no dateutil then doing the git tango is pointless.
    def run_git_log():
        pass


def get_pkg_version():
    """Get version string by parsing PKG-INFO."""
    try:
        with open("PKG-INFO", "r") as fp:
            rgx = re.compile(r"Version: (\d+)")
            for line in fp.readlines():
                match = rgx.match(line)
                if match:
                    return match.group(1)
    except IOError:
        return None


def get_version():
    """Tries to get the EFILTER version from git or PKG-INFO.

    The EFILTER version is the UTC UNIX epoch of latest git commit.

    Example:
        1438623992
    """
    version = run_git_log()
    if version:
        return version

    version = get_pkg_version()
    if version:
        return version

    raise RuntimeError("Couldn't get git log or PKG-INFO to guess version.")



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER query normalizer.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


from efilter import dispatch
from efilter import ast
from efilter import query as q


@dispatch.multimethod
def normalize(expr):
    """Optimizes the AST for better performance and simpler structure.

    The returned query will be logically equivalent to what was provided but
    transformations will be made to flatten and optimize the structure. This
    works by recognizing certain patterns and replacing them with nicer ones,
    eliminating pointless expressions, and so on.

    Examples:
        # Logical expressions are made variadic:
        Intersection("foo", Intersection("bar", "baz")) # becomes:
        Intersection("foo", "bar", "baz")

        # Map-forms are rotated so that the LHS is a Binding when possible:
        Map(
            Map(
                Binding("Process"),
                Binding("parent")),
            Equivalence(
                Binding("name"),
                Literal("init")))
        # Becomes:
        Map(
            Binding("Process"),
            Map(
                Binding("parent"),
                Equivalence(
                    Binding("name"),
                    Literal("init"))))
    """
    _ = expr
    raise NotImplementedError()


@normalize.implementation(for_type=q.Query)
def normalize(query):
    new_root = normalize(query.root)
    return q.Query(query, root=new_root)


@normalize.implementation(for_type=ast.Expression)
def normalize(expr):
    return expr


@normalize.implementation(for_type=ast.Map)
def normalize(expr):
    """Rotate nested map-forms so they cascade on the RHS.

    Basic map-forms should be rotated as follows:
    (map (map x y) (...)) => (map x (map y) (...))

    These are functionally equivalent, but the latter is easier to follow.

    Returns rotated Map instance.
    """
    lhs = normalize(expr.lhs)
    rhs = normalize(expr.rhs)

    if (isinstance(lhs, ast.Map)
            and isinstance(lhs.lhs, ast.Binding)):
        lhs_ = lhs.lhs
        rhs = type(expr)(lhs.rhs, rhs, start=lhs.rhs.start, end=rhs.end)
        lhs = lhs_

    return type(expr)(lhs, rhs, start=lhs.start, end=rhs.end)


@normalize.implementation(for_type=ast.Within)
def normalize(expr):
    """any, each, filter and sort are not cascaded.

    This is basically a pass-through function.
    """
    lhs = normalize(expr.lhs)
    rhs = normalize(expr.rhs)
    return type(expr)(lhs, rhs, start=lhs.start, end=rhs.end)


@normalize.implementation(for_type=ast.BinaryExpression)
def normalize(expr):
    """Eliminate if either of the children is None."""
    lhs = normalize(expr.lhs)
    rhs = normalize(expr.rhs)

    if lhs is None:
        if rhs is None:
            return None

        return rhs
    elif rhs is None:
        return lhs

    return type(expr)(lhs, rhs, start=lhs.start, end=rhs.end)


@normalize.implementation(for_type=ast.VariadicExpression)
def normalize(expr):
    """Pass through n-ary expressions, and eliminate empty branches.

    Variadic and binary expressions recursively visit all their children.

    If all children are eliminated then the parent expression is also
    eliminated:

    (& [removed] [removed]) => [removed]

    If only one child is left, it is promoted to replace the parent node:

    (& True) => True
    """
    children = []
    for child in expr.children:
        branch = normalize(child)
        if branch is None:
            continue

        if type(branch) is type(expr):
            children.extend(branch.children)
        else:
            children.append(branch)

    if len(children) == 0:
        return None

    if len(children) == 1:
        return children[0]

    return type(expr)(*children, start=children[0].start,
                      end=children[-1].end)



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system.

This special protocol defines functions for an application/service level
delegate object, intended to provide type information and a global name service
to EFILTER. Implementing this protocol will enable EFILTER expressions to
reference globals and provide stronger type protections and hints in the
query analyzer.
"""

from efilter import dispatch
from efilter import protocol


# pylint: disable=unused-argument


def reflect(reflective_type, name):
    """Provide the type of 'name', which is either a global or part of 'scope'.

    Arguments:
        delegate: The application delegate.
        name: The name to be reflected. Either a global (from getnames) or
            a member of 'scope', which is a type or a container.
        scope: The optional type scope, which was returned by a previous call
            to 'reflect'. Used to qualify that 'name' is not global, but a
            member of 'scope'.

    Returns:
        The type of 'name' or None. Invalid names should return None,
        whereas valid names with unknown type should return AnyType.

    Examples:
        # What's a process?
        reflect(delegate, "Process") #=> Process

        # What's pslist?
        reflect(delegate, "pslist") #=> <Plugin pslist>

        # What's the pid column's type?
        reflect(
            reflect(delegate, "pslist"),
            "pid") #=> int
    """
    raise NotImplementedError()


def _liberal_dispatch(args, kwargs):
    """This makes the multimethod accept both classes and their instances.

    You may be wondering why. It's a fair question. Most other protocols in
    efilter use the default multimethod dispatch, which decides which overload
    to call based on the type of the first argument. The first version of the
    reflective protocol used a dispatch that worked on the types themselves,
    instead of instances: you would pass the multimethods the actual type as
    first argument, instead of an object, and this worked well with class
    methods and all was well.

    Enter Rekall. Rekall, because it simulates the type systems of a large
    number of different operating system versions, doesn't know what type a
    particular object will be until runtime, when it can determine the correct
    types based on debug symbols from the OS being analyzed. There was no
    reasonable way for Rekall to use this protocol! Simultaneously, the vast
    majority of other projects will be perfectly well served by class-level
    implementations, and a runtime reflection system would be burdensome.

    The most practical solution is what I here refer to as liberal dispatch -
    if you pass a multimethod with liberal dispatch a type, it will use it. If
    not, it will dispatch on the type of the instance. Everyone wins.

    Direct your hate mail (or better ideas) to adamsh@google.com.
    """

    if not args:
        raise ValueError(
            "Multimethods must be passed at least one positional arg.")

    if isinstance(args[0], type):
        return args[0]
    else:
        return type(args[0])

    return args[0]


reflect = dispatch.multimethod(reflect,
                               dispatch_function=_liberal_dispatch)


def getkeys(reflective_type):
    """Provide a list of keys that can be reflected or selected/resolved."""
    raise NotImplementedError()


getkeys = dispatch.multimethod(getkeys,
                               dispatch_function=_liberal_dispatch)


class IReflective(protocol.Protocol):
    _protocol_functions = (reflect, getkeys)


IReflective.implement(for_type=dict,
                      implementations={
                          reflect: lambda _, __: protocol.AnyType,
                          getkeys: lambda d: d.keys()})


IReflective.implement(for_type=protocol.AnyType,
                      implementations={
                          reflect: lambda _, __: protocol.AnyType,
                          getkeys: lambda _: ()})



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER Abstract Syntax Tree.

The AST represents the actual canonical syntax of EFILTER, as understood by all
the behavior implementations and transformations. The string and lisp-based
syntaxes are frontends that translate into this AST, which is what is actually
interpretted.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import protocol

from efilter.protocols import associative
from efilter.protocols import boolean
from efilter.protocols import eq
from efilter.protocols import iset
from efilter.protocols import ordered
from efilter.protocols import number


class Expression(object):
    """Base class of the query AST.

    Behavior of the query language is encoded in the various transform
    functions. Expression themselves have no behavior, and only contain
    children and type and arity information.
    """

    __abstract = True

    children = ()
    arity = 0
    start = None
    end = None

    type_signature = (protocol.AnyType,)
    return_signature = protocol.AnyType

    def __hash__(self):
        return hash((type(self), self.children))

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.children == other.children

    def __ne__(self, other):
        return not self.__eq__(other)

    def __init__(self, *children, **kwargs):
        super(Expression, self).__init__()

        self.start = kwargs.pop("start", None)
        self.end = kwargs.pop("end", None)

        if kwargs:
            raise ValueError("Unexpected argument(s) %s" % kwargs.keys())

        if self.arity and len(children) != self.arity:
            raise ValueError("%d-ary expression %s passed %d children." % (
                self.arity, type(self).__name__, len(children)))

        self.children = children

    def __repr__(self):
        if len(self.children) == 1:
            return "%s(%r)" % (type(self).__name__, self.children[0])

        lines = []
        for child in self.children:
            if isinstance(child, Expression):
                clines = [" %s" % line for line in repr(child).split("\n")]
            else:
                clines = repr(child).split("\n")
            lines.extend(clines)

        return "%s(\n%s)" % (type(self).__name__, "\n".join(lines))


class ValueExpression(Expression):
    """Unary expression."""
    arity = 1
    __abstract = True

    @property
    def value(self):
        return self.children[0]


class BinaryExpression(Expression):
    arity = 2
    __abstract = True

    @property
    def lhs(self):
        return self.children[0]

    @property
    def rhs(self):
        return self.children[1]


class VariadicExpression(Expression):
    """Represents an expression with variable arity."""

    arity = None
    __abstract = True


### Value (unary) expressions ###

class Literal(ValueExpression):
    """Represents a literal, which is to say not-an-expression."""

    type_signature = None  # Depends on literal.


class Binding(ValueExpression):
    """Represents a member of the evaluated object - attributes of entity."""

    type_signature = (associative.IAssociative,)


class Complement(ValueExpression):
    """Logical NOT."""

    type_signature = (boolean.IBoolean,)
    return_signature = boolean.IBoolean


### Binary expressions ###


class IsInstance(BinaryExpression):
    """Evaluates to True if the current scope is an instance of type."""


class Within(BinaryExpression):
    """Uses left side as new bindings and evaluates right side as a subquery.

    Concrete behavior depends on the various subclasses, such as Filter and
    Map, but each one of them will expect left hand side to be an associative
    object holding the new bindings, or a repeated variable of associative
    objects.
    """
    __abstract = True
    type_signature = (associative.IAssociative, protocol.AnyType)
    return_signature = None  # Depends on RHS.

    @property
    def context(self):
        return self.lhs

    @property
    def expression(self):
        return self.rhs


class Map(Within):
    """Returns the result of applying right side to the values on left side.

    If left is a repeated value then this will return another repeated value.
    """


class Filter(Within):
    """Filters (repeated) values on left side using expression on right side.

    Will return a repeated variable containing only the values for which the
    expression on the right evaluated to true.
    """


class Sort(Within):
    """Sorts the left hand side using the right hand side return."""


class Any(Within):
    """Returns true if the rhs evaluates as true for any value of lhs."""
    return_signature = bool


class Each(Within):
    """Returns true if the rhs evaluates as true for every value of lhs."""
    return_signature = bool


class Membership(BinaryExpression):
    """Membership of element in set."""
    type_signature = (eq.IEq, iset.ISet)
    return_signature = boolean.IBoolean

    @property
    def element(self):
        return self.lhs

    @property
    def set(self):
        return self.rhs


class RegexFilter(BinaryExpression):
    type_signature = (basestring, basestring)
    return_signature = boolean.IBoolean

    @property
    def string(self):
        return self.lhs

    @property
    def regex(self):
        return self.rhs


### Variadic Expressions ###

### Logical Variadic ###

class LogicalOperation(VariadicExpression):
    type_signature = boolean.IBoolean
    return_signature = boolean.IBoolean
    __abstract = True


class Union(LogicalOperation):
    """Logical OR (variadic)."""


class Intersection(LogicalOperation):
    """Logical AND (variadic)."""


### Variadic Relations ###

class Relation(VariadicExpression):
    return_signature = boolean.IBoolean
    __abstract = True


class OrderedSet(Relation):
    """Abstract class to represent strict and non-strict ordering."""

    type_signature = ordered.IOrdered
    __abstract = True


class StrictOrderedSet(OrderedSet):
    """Greater than relation."""

    type_signature = ordered.IOrdered


class PartialOrderedSet(OrderedSet):
    """Great-or-equal than relation."""

    type_signature = ordered.IOrdered


class ContainmentOrder(Relation):
    """Inclusion of set 1 by set 2 and so on."""

    type_signature = iset.ISet


class Equivalence(Relation):
    """Logical == (variadic)."""

    type_signature = eq.IEq


### Variadic Arithmetic ###

class NumericExpression(VariadicExpression):
    """Arithmetic expressions."""

    return_signature = number.INumber
    __abstract = True


class Sum(NumericExpression):
    """Arithmetic + (variadic)."""

    type_signature = number.INumber


class Difference(NumericExpression):
    """Arithmetic - (variadic)."""

    type_signature = number.INumber


class Product(NumericExpression):
    """Arithmetic * (variadic)."""

    type_signature = number.INumber


class Quotient(NumericExpression):
    """Arithmetic / (variadic)."""

    type_signature = number.INumber



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system.

The repeated protocol concerns itself with variables that have more than one
value, such as repeated fields on protocol buffers; superpositions (see
protocols.superposition) are a special case of repeated variables.
"""

from efilter import dispatch
from efilter import protocol

from efilter.protocols import eq

# Declarations:
# pylint: disable=unused-argument


@dispatch.multimethod
def repeated(first_value, *values):
    """Build a repeated variable from values, all of which are the same type.

    Repeated values preserve order and allow the same value to appear more
    than once.
    """
    raise NotImplementedError()


def meld(*values):
    """Return the repeated value, or the first value if there's only one.

    This is a convenience function, equivalent to calling
    getvalue(repeated(x)) to get x.

    This function skips over instances of None in values (None is not allowed
    in repeated variables).

    Examples:
        meld("foo", "bar") # => ListRepetition("foo", "bar")
        meld("foo", "foo") # => ListRepetition("foo", "foo")
        meld("foo", None) # => "foo"
        meld(None) # => None
    """
    values = [x for x in values if x is not None]
    if not values:
        return None

    result = repeated(*values)
    if isrepeating(result):
        return result

    return getvalue(result)


@dispatch.multimethod
def getvalues(x):
    """Return a collection of the values of x."""
    raise NotImplementedError()


def getvalue(x):
    """Return the single value of x or raise TypError if more than one value."""
    if isrepeating(x):
        raise TypeError(
            "Ambiguous call to getvalue for %r which has more than one value."
            % x)

    for value in getvalues(x):
        return value


@dispatch.multimethod
def value_type(x):
    """Return the type (class) of the values of x."""
    raise NotImplementedError()


@dispatch.multimethod
def value_eq(x, y):
    """Are the values of x and y exactly the same?"""
    raise NotImplementedError()


@dispatch.multimethod
def value_apply(x, f):
    """Apply f to each value of x and return a new repeated var of results."""
    raise NotImplementedError()


@dispatch.multimethod
def isrepeating(x):
    """Optional: Is x a repeated var AND does it have more than one value?"""
    return isinstance(x, IRepeated) and len(getvalues(x)) > 1


class IRepeated(protocol.Protocol):
    _protocol_functions = (getvalues, value_type, value_eq, value_apply)


def _scalar_value_eq(x, y):
    if isrepeating(y):
        return False

    return eq.eq(x, getvalue(y))


# Implementation for scalars:
# pylint: disable=unnecessary-lambda
IRepeated.implement(
    for_type=protocol.AnyType,
    implementations={
        getvalues: lambda x: (x,) if x is not None else (),
        value_type: lambda x: type(x),
        value_eq: _scalar_value_eq,
        value_apply: lambda x, f: f(x)
    }
)
